﻿using System;
using System.IO;
using System.IO.IsolatedStorage;
using System.Linq;
using System.Threading;
using System.Windows.Forms;

namespace CanonCopier
{
    public partial class MainForm : Form
    {
        delegate DriveInfo GetSelectedDriveEvent();
        delegate void BasicEvent();
        delegate void SetProgressBarMaximumEvent(int max);
        delegate void SetStatusTextEvent(string text);
		/// <summary>
		/// Maximum number of characters to display on the status text when showing which
		/// destination folder the files will be moved to.
		/// </summary>
		private const int DestinationFolderDisplayLimit = 40;
		/// <summary>
		/// Flag that determines whether the moving operation had begun.
		/// </summary>
        private bool IsMoving = false;
		/// <summary>
		/// Flag that determines if the moving operation is being cancelled.
		/// </summary>
        private bool IsCancelling = false;
		/// <summary>
		/// The path to move files to when no path is set.
		/// </summary>
		private static readonly string DefaultDestinationFolder = Environment.GetFolderPath(Environment.SpecialFolder.MyPictures);
		private string _DestinationFolder;
		/// <summary>
		/// The path to move files to.
		/// </summary>
		public string DestinationFolder
		{
			get { return _DestinationFolder; }
			set
			{
				_DestinationFolder = value;
				ShowDestinationFolder();
			}
		}
		/// <summary>
		/// Shows the destination folder on the status text.
		/// </summary>
		private void ShowDestinationFolder()
		{
			StatusText.Text = "Files will be moved to ";

			var pathLength = DestinationFolder.Length;

			/// truncate the path if the path is longer than the display limit
			if (pathLength > DestinationFolderDisplayLimit)
			{
				StatusText.Text += DestinationFolder.Substring(0, DestinationFolderDisplayLimit / 2) +
					"..." + DestinationFolder.Substring(pathLength - DestinationFolderDisplayLimit / 2,
					DestinationFolderDisplayLimit / 2);
			}
			else StatusText.Text += DestinationFolder;
		}

        public MainForm()
        {
            InitializeComponent();
        }

        private void MainForm_Load(object sender, EventArgs e)
        {
			LoadSettings();

			/// show all removable drives found
            DriveSelect.DisplayMember = "VolumeLabel";

            foreach (var drive in Directory.GetLogicalDrives())
            {
                var info = new DriveInfo(drive);
                if (info.DriveType == DriveType.Removable)
                {
                    DriveSelect.Items.Add(info);
                }
            }

			/// select the first removable drive if available; disable
			/// the Go button otherwise
			if (DriveSelect.Items.Count > 0) DriveSelect.SelectedIndex = 0;
			else GoButton.Enabled = false;
        }
		/// <summary>
		/// Loads settings from the setting file.
		/// </summary>
		private void LoadSettings()
		{
			string content = null;
			StreamReader reader = null;

			try
			{
				var store = IsolatedStorageFile.GetUserStoreForAssembly();
				reader = new StreamReader(new IsolatedStorageFileStream("settings.txt", FileMode.Open, store));
				content = reader.ReadToEnd();
			}
			catch { }
			finally
			{
				if (reader != null) reader.Dispose();
			}

			/// default the destination folder if settings is either empty
			/// or null
			DestinationFolder = DefaultDestinationFolder;
			if (content != null)
			{
				var segments = content.Split(new char[] { ':' }, 2);
				if (segments[0] == "DestinationFolder") DestinationFolder = segments[1];
			}
		}
		/// <summary>
		/// Fires when the Go button is clicked.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
        private void GoButton_Click(object sender, EventArgs e)
        {
            if (IsMoving)
            {
                IsCancelling = true;
            }
            else
            {
                var thread = new Thread(new ThreadStart(MoveFiles));
                thread.Start();

                GoButton.Text = "Cancel";
                IsMoving = true;
            }
        }
		/// <summary>
		/// Does the move file oepration.
		/// </summary>
        private void MoveFiles()
        {
			/// whether a multithread safe approach to calling methods in the main form
			/// will be determined by this one control's requirement
            var invokeRequired = DriveSelect.InvokeRequired;

			/// get the drive selected
            var drive = invokeRequired ?
                (DriveInfo)this.Invoke(new GetSelectedDriveEvent(GetSelectedDrive)) :
                GetSelectedDrive();

            var path = Path.Combine(drive.Name, "DCIM");
            var fileCount = 0;
            var movedCount = 0;

            if (Directory.Exists(path))
            {
				/// find out how many files need to be moved
                foreach (var directory in new DirectoryInfo(path).GetDirectories())
                {
                    fileCount += directory.GetFiles().Count();
                }

				EnsureDestinationFolder();

				/// set up the progress bar and hide the change link
				if (invokeRequired)
				{
					this.Invoke(new SetProgressBarMaximumEvent(SetProgressBarMaximum), new object[] { fileCount });
					this.Invoke(new BasicEvent(ToggleChangeLinkVisible));
				}
				else
				{
					SetProgressBarMaximum(fileCount);
					ToggleChangeLinkVisible();
				}

				/// for each subdirectory, move the files over
                foreach (var directory in new DirectoryInfo(path).GetDirectories())
                {
                    foreach (var file in directory.GetFiles())
                    {
						/// break from the loop if the operation is cancelled
                        if (IsCancelling) goto BreakLoop;

                        string statusText = "Moving " + file.Name;
                        if (invokeRequired) this.Invoke(new SetStatusTextEvent(SetStatusText), new object[] { statusText });
                        else SetStatusText(statusText);

                        try
                        {
                            file.MoveTo(Path.Combine(DestinationFolder, file.Name));
                        }
                        catch (Exception ex)
                        {
                            MessageBox.Show("Error moving file: " + ex.Message,
                                "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                        }

                        ++movedCount;

						if (invokeRequired) this.Invoke(new BasicEvent(ProgressBarPerformStep));
						else ProgressBarPerformStep();
                    }
                }

            BreakLoop:

                MessageBox.Show(movedCount + " files moved.", "Success", MessageBoxButtons.OK, MessageBoxIcon.Information);
            }
            else
            {
                MessageBox.Show("Drive does not contain Canon directories.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Warning);
            }

			/// reset the form once everything is done
            if (invokeRequired) this.Invoke(new BasicEvent(FormReset));
            else FormReset();
        }
		/// <summary>
		/// Ensures that the destination folder exists.
		/// </summary>
		private void EnsureDestinationFolder()
		{
			if (!Directory.Exists(DestinationFolder))
			{
				var folders = DestinationFolder.Split(Path.PathSeparator);

				for (var i = 0; i < folders.Length; ++i)
				{
					var path = string.Join(Path.PathSeparator.ToString(), folders, 0, i + 1);
					if (!Directory.Exists(path)) Directory.CreateDirectory(path);
				}
			}
		}
		/// <summary>
		/// Resets the form.
		/// </summary>
        private void FormReset()
        {
            IsCancelling = false;
            IsMoving = false;
            ProgressBar.Value = 0;
			ShowDestinationFolder();
            GoButton.Text = "Go!";
			ToggleChangeLinkVisible();
        }
		/// <summary>
		/// Gets the drive selected.
		/// </summary>
		/// <returns></returns>
        private DriveInfo GetSelectedDrive()
        {
            return DriveSelect.SelectedItem as DriveInfo;
        }
		/// <summary>
		/// Sets the status text.
		/// </summary>
		/// <param name="text"></param>
        private void SetStatusText(string text)
        {
            StatusText.Text = text;
        }
		/// <summary>
		/// Sets the maximum value for the progress bar.
		/// </summary>
		/// <param name="max"></param>
        private void SetProgressBarMaximum(int max)
        {
            ProgressBar.Maximum = max;
        }
		/// <summary>
		/// Performs a step on the progress bar so it moves forward.
		/// </summary>
        private void ProgressBarPerformStep()
        {
            ProgressBar.PerformStep();
        }
		/// <summary>
		/// Resets the progress bar.
		/// </summary>
        private void ProgressBarReset()
        {
            ProgressBar.Value = 0;
        }
		private void ToggleChangeLinkVisible()
		{
			LinkChange.Visible = !LinkChange.Visible;
		}
		/// <summary>
		/// Fires when user selects a different destination folder to move files to.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void LinkChange_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
		{
			var dialog = new FolderBrowserDialog();
			dialog.ShowNewFolderButton = true;
			dialog.Description = "Select a destination folder";
			var result = dialog.ShowDialog(this);

			if (result == DialogResult.OK)
			{
				DestinationFolder = dialog.SelectedPath;
				SaveSettings();
			}
		}
		/// <summary>
		/// Saves the program settings to a file.
		/// </summary>
		private void SaveSettings()
		{
			var store = IsolatedStorageFile.GetUserStoreForAssembly();
			var writer = new StreamWriter(new IsolatedStorageFileStream("settings.txt", FileMode.Create, store));
			writer.Write("DestinationFolder:");
			writer.Write(DestinationFolder);
			writer.Dispose();
		}
    }
}
